use tempdb
go

/*************************************************************************
 *****			sp_get_sparse_information version 0.9               ******
 *****			  author: Qingsong Yao								******
 *****	 get the column's notnull count from a table,				******
 *****	    also determine whether it can be sparse column          ******
 *****	  usage:													******
 *****			  exec tempdb.dbo.sp_get_sparse_information			******
 *****						@dbname='yourdb',						******
 *****						@tablename='tablename',					******
 *****						@samplerate= samplerate,				******
 *****						@adjustratio =adjustratio				******
 *****		samplerate is the ratio (1 to 100) of the source table	******
 *****		the procedure use the cutoffrate specified in BOL, and  ******
 *****		it will use the nullpercentage of a column compare the  ******
 *****			@adjustratio * cutoffrate of a column type to		******
 *****		determine whether a column can be spase or not.			******
 *************************************************************************/ 
 
if object_id('sp_get_sparse_information') is not null
  drop procedure sp_get_sparse_information
go

create procedure sp_get_sparse_information (@dbname as nvarchar(128),@tablename as nvarchar(128), @samplerate int =100, @adjustratio numeric(3,2)=0.8)
as
begin
set nocount on

-- 1. set schema.object as schema name, object as table name
declare @schematablename nvarchar(200) = QUOTENAME(@dbname)+'.' + @tablename
declare @location int = charindex('.',@tablename,1)

if (@location <=0)
  set @schematablename = QUOTENAME(@dbname)+'.dbo.' + QUOTENAME(@tablename)  
else
  set @tablename = SUBSTRING(@tablename, @location+1,LEN(@tablename)-@location)
declare @objectid int = (select OBJECT_ID(@schematablename))
--2. clean the tempory tables
if object_id('tempcolumninfo') is not null 
	drop table tempcolumninfo
if object_id('sparse_info_tempcount') is not null 
	drop table sparse_info_tempcount
declare @table_nullcount varchar(200)=QUOTENAME(@tablename+'_nullcount')
declare @statement nvarchar(max)
select @statement ='; if object_id('''+ @table_nullcount +''') is not null drop table ' + @table_nullcount
exec (@statement)

--populate column information from target database into tempary table
select @statement  ='use ' + QUOTENAME(@dbname) 
select @statement += '; select columns.*, types.name as typename,TYPE_NAME(types.system_type_id) as systemtypename,is_assembly_type '
select @statement +=' into tempdb.dbo.tempcolumninfo from sys.columns '
select @statement +=' join sys.types on columns.user_type_id = types.user_type_id where object_id = object_id(''' +@schematablename +''')'
exec (@statement)

--this statement generate a widetable with notnull count for all columns, for text, image, ntext we need to convert them into large value type forst
select @statement  =' select count(1) as mymyrowcount,'
select @statement +='  count(     ['+name+'])					  as ['+name+ '],' from tempcolumninfo where system_type_id not in (99,35,34)
select @statement +='  count(cast(['+name+'] as nvarchar(max)))  as ['+name+ '],' from tempcolumninfo where system_type_id =99
select @statement +='  count(cast(['+name+'] as  varchar(max)))  as ['+name+ '],' from tempcolumninfo where system_type_id =35
select @statement +='  count(cast(['+name+'] as varbinary(max))) as ['+name+ '],' from tempcolumninfo where system_type_id =34
select @statement  = substring(@statement,1,len(@statement)-1)
select @statement +=' into sparse_info_tempcount from ' + @schematablename + ' TableSample ( ' + cast(@samplerate as nvarchar(10)) +' Percent)'
exec (@statement)

--using unpoivot command to convert columns/notnullcount into rows and join with tempcolumninfo, insert the result into nullcount table
declare @columns nvarchar(max)=''
select  @columns    = @columns + '[' + name +'],' from tempdb.dbo.tempcolumninfo
select  @columns    = substring(@columns,1,len(@columns)-1) 
select @statement  ='declare @cnt numeric(20,0) = (select mymyrowcount from sparse_info_tempcount) '
select @statement +=' if (@cnt =0) set @cnt =1 '
select @statement +=' select ColumnName, NotNullCount, + (@cnt - NotNullCount) as NullCount,(1.0 - NotNullCount/@cnt) *100 as NullPercentage, '
select @statement +=' cast(''possible'' as varchar(40)) as canbe_sparse, cast('''' as varchar(400)) as reason, '
select @statement +=' cast('' alter table ' + @schematablename + ' alter column ['' + ColumnName +''] add sparse '' as nvarchar(255)) as addsparse_statement,'
select @statement +=' cast('' alter table ' + @schematablename + ' alter column ['' + ColumnName +''] drop sparse '' as nvarchar(255)) as dropsparse_statement,'
select @statement +=' column_id,typename, systemtypename, max_length,precision,scale,is_nullable,is_rowguidcol,is_assembly_type,'
select @statement +=' is_identity,is_computed,is_filestream,default_object_id,rule_object_id,is_sparse,is_column_set,system_type_id,user_type_id'
select @statement +=' into ' + @table_nullcount 
select @statement +=' from (select * from sparse_info_tempcount) as p'
select @statement +=' unpivot (NotNullCount for ColumnName in ( ' + @columns+')) as s '
select @statement +=' join tempdb.dbo.tempcolumninfo on s.ColumnName =tempcolumninfo.Name  collate latin1_general_ci_as'
exec (@statement)

 -- update the canbe_sparse and reason column
 select @statement ='update ' + @table_nullcount +' set canbe_sparse = ''false'', reason =reason+''; column is always not null'' where is_nullable =0'
 exec (@statement)
 select @statement ='update ' + @table_nullcount +' set canbe_sparse = ''false'', reason =reason+''; column is rowguid'' where is_rowguidcol =1'
 exec (@statement)
 select @statement ='update ' + @table_nullcount +' set canbe_sparse = ''false'', reason =reason+''; column is identity'' where is_identity =1'
 exec (@statement)
 select @statement ='update ' + @table_nullcount +' set canbe_sparse = ''false'', reason =reason+''; column is computed'' where is_computed =1'
 exec (@statement)
 select @statement ='update ' + @table_nullcount +' set canbe_sparse = ''false'', reason =reason+''; column is filestream'' where is_filestream =1'
 exec (@statement)
 select @statement ='update ' + @table_nullcount +' set canbe_sparse = ''false'', reason =reason+''; column has default'' where default_object_id !=0'
 exec (@statement)
 select @statement ='update ' + @table_nullcount +' set canbe_sparse = ''false'', reason =reason+''; column has rule'' where rule_object_id !=0'
 exec (@statement)
 select @statement ='update ' + @table_nullcount +' set canbe_sparse = ''false'', reason =reason+''; column is columnset'' where is_column_set =1'
 exec (@statement)
 select @statement ='update ' + @table_nullcount +' set canbe_sparse = ''false'', reason =reason+''; column is CLR UDT'' where is_assembly_type =1'
 exec (@statement)
 select @statement ='update ' + @table_nullcount +' set canbe_sparse = ''change type'', reason =reason+''; column is text/image/ntext type'' where system_type_id in (99,35,34)'
 exec (@statement)
  --primary key/clustered index
 select @statement ='update ' + @table_nullcount +' set canbe_sparse = ''false'',reason =reason+''; column is in clustered index'' where column_id in ('
 select @statement +='select column_id from '+QUOTENAME(@dbname)+'.sys.index_columns where object_id='+cast(@objectid as varchar(20))+' and index_id=1)'
 exec (@statement)
 --non-clustered index and not partition_key
 select @statement ='update ' + @table_nullcount +' set canbe_sparse = ''false'',reason =reason+''; column is in non-clustered index'' '
 select @statement +=' where column_id in (select column_id from '+QUOTENAME(@dbname)+'.sys.index_columns '
 select @statement  +=' where object_id='+cast(@objectid as varchar(20))+' and index_id>1 and partition_ordinal =0)'
 exec (@statement)
 
 --non-clustered index and partition_key
 select @statement ='update ' + @table_nullcount +' set canbe_sparse = ''false'',reason =reason+''; column is in partition key'' '
 select @statement +=' where column_id in (select column_id from '+QUOTENAME(@dbname)+'.sys.index_columns '
 select @statement  +=' where object_id='+cast(@objectid as varchar(20))+' and index_id>1 and partition_ordinal >0)'
 exec (@statement)
 
 --unique constaints or unquenon-clustered index and partition_key
 select @statement ='update ' + @table_nullcount +' set canbe_sparse = ''false'',reason =reason+''; column is in unique or in unique index'' '
 select @statement +=' where column_id in (select column_id from '+QUOTENAME(@dbname)+'.sys.index_columns as a join ' 
 select @statement += QUOTENAME(@dbname) +'.sys.indexes as b on a.object_id = b.object_id and a.index_id = b.index_id '
 select @statement  +=' where a.object_id='+cast(@objectid as varchar(20))+' and (is_unique =1 or is_unique_constraint=1))'
 exec (@statement)
      
 select @statement ='declare @tempratio table (datatypename nvarchar(20), cutoffratio numeric(3,2))'
 select @statement +='insert into @tempratio values '
 select @statement += '(''tinyint'',0.86),(''smallint'',0.76),(''int'',0.64),(''bigint'',0.52),(''real'',0.64),(''float'',0.52),'
 select @statement +='(''smallmoney'',0.64),(''money'',0.52),(''decimal'',0.55),(''numeric'', 0.55),'
 select @statement +='(''smalldatetime'',0.64),(''datetime'',0.52),(''date'',0.69),(''datetime2'',0.57),(''time'',0.69),(''datetimetoffset(0)'',0.52),'
 select @statement +='(''varchar'',0.6),(''char'',0.6),(''nvarchar'',0.6),(''nchar'',0.6),(''varbinary'',0.6),(''binary'',0.6),'
 select @statement +='(''sql_variant'',0.50),(''xml'',0.6),(''hierarchyId'',0.6),(''uniqueidentifier'',0.43) '
 select @statement +='update ' + @table_nullcount +' set canbe_sparse = ''false'',reason =reason+''; null precentage is smaller than predefine ratio'' '
 select @statement +=' where column_id in (select column_id from ' + @table_nullcount +' as a join @tempratio on a.systemtypename = datatypename collate latin1_general_ci_as'
 select @statement +=' where a.NullPercentage < (cutoffratio * 100.0 * ' + CAST(@adjustratio as nvarchar(10)) +'))'
 exec (@statement)

 -- display information
 PRINT 'finish populating sparse information, run query: select * from [' + @tablename +'_nullcount] to get the result'
 set @statement ='select ColumnName,TypeName, NotNullCount,NullCount, NullPercentage,canbe_sparse,reason,addsparse_statement,dropsparse_statement from [' + @tablename +'_nullcount] order by ColumnName;'
 exec (@statement)
end
go

--sample statement
exec tempdb.dbo.sp_get_sparse_information  @dbname='spatial_with_filestream2', 
		@tablename='dbo.photo',
		@samplerate=100,
		@adjustratio=0.8





