

-- find all the columns in sparse columnset
select * from sys.columns
where object_id = OBJECT_ID('maxcols')

-- create xml schema from metadata
declare @t table (thexml xml)
declare @stmt varchar(max)

-- xml path concatenation trick
set @stmt = 
(select ',' + name 
from sys.columns
where object_id = OBJECT_ID('maxcols')
and is_sparse = 1
for xml path(''))

-- concatenate rest of select for xml xmlschema statement
set @stmt = 'select top(0) ' + STUFF(@stmt, 1, 1, '') 
  + ' from maxcols for xml raw(''columnset''), elements, type, xmlschema(''http://foo'')'

-- execute it into a table variable
insert into @t
exec(@stmt)

-- transform to 
--  a. get rid of targetNamespace
--  b. get rid of containing columnset tags
declare @x xml
set @x = (
select top(1) thexml.query('
declare namespace xsd = "http://www.w3.org/2001/XMLSchema";
<xsd:schema xmlns:sqltypes = "http://schemas.microsoft.com/sqlserver/2004/sqltypes">
<xsd:import namespace="http://schemas.microsoft.com/sqlserver/2004/sqltypes" />
{
/xsd:schema/xsd:element/xsd:complexType/xsd:sequence/xsd:element
}
</xsd:schema>
')
from @t)

-- get rid of all minOccurs on global element declaration
set @x.modify('delete //@minOccurs')
-- to check it
select @x

-- create schema collection from variable
create xml schema collection sparse_schema as @x
go

declare @x xml(sparse_schema)
set @x = '<col1>12</col1><col2>12</col2>'
go

declare @x xml(sparse_schema)
set @x = '<col1>12</col1><col2>foo</col2>'
go

-- limitation, you can include more than one of same element
-- provides only type-checking
declare @x xml(sparse_schema)
set @x = '<col1>12</col1><col1>1</col1>'
go

/* original

<xsd:schema targetNamespace="http://foo" xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:sqltypes="http://schemas.microsoft.com/sqlserver/2004/sqltypes" elementFormDefault="qualified">
  <xsd:import namespace="http://schemas.microsoft.com/sqlserver/2004/sqltypes" schemaLocation="http://schemas.microsoft.com/sqlserver/2004/sqltypes/sqltypes.xsd" />
  <xsd:element name="columnset">
    <xsd:complexType>
      <xsd:sequence>
        <xsd:element name="col1" type="sqltypes:int" minOccurs="0" />
        <xsd:element name="col2" type="sqltypes:int" minOccurs="0" />
        <xsd:element name="col3" type="sqltypes:int" minOccurs="0" />
        <xsd:element name="col4" type="sqltypes:int" minOccurs="0" />
        <xsd:element name="col5" type="sqltypes:int" minOccurs="0" />
        <xsd:element name="col6" type="sqltypes:int" minOccurs="0" />
        <xsd:element name="col7" type="sqltypes:int" minOccurs="0" />
        <xsd:element name="col8" type="sqltypes:int" minOccurs="0" />
        <xsd:element name="col9" type="sqltypes:int" minOccurs="0" />
        <xsd:element name="col10" type="sqltypes:int" minOccurs="0" />
      </xsd:sequence>
    </xsd:complexType>
  </xsd:element>
</xsd:schema>

*/